//  test_selection.cpp
//
//  Copyright 2007 University of Pittsburgh.
//  U of Pittsburgh coding done by Brady Hunsaker and Mustafa Baz.
//
//  This file is part of STOP (Selection Tool for Optimization Parameters).
//
//  The contents of this file are subject to the Common Public License
//  1.0.  (the "License"); you may not use this file except in
//  compliance with the License. You should have received a copy of
//  the Common Public License along with STOP.
//
//  Software distributed under the License is distributed on an "AS
//  IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
//  implied. See the License for the specific language governing
//  rights and limitations under the License.
//
//  Alternatively, the contents of this file may be used under the
//  terms of the GNU General Public License Version 2 or later (the
//  "GPL"), in which case the provisions of the GPL are applicable
//  instead of those above. If you wish to allow use of your version
//  of this file only under the terms of the GPL, and not to allow
//  others to use your version of this file under the terms of the
//  CPL, indicate your decision by deleting the provisions above and
//  replace them with the notice and other provisions required by the
//  GPL. If you do not delete the provisions above, a recipient may
//  use your version of this file under the terms of either the CPL or
//  the GPL.
//
/////////////////////////////////////////////////////////////
//  These routines offer different methods to select a set
//  of test instances.

//#include <vector>
//#include "MersenneTwister.h"
#include "test_selection.hpp"

using namespace std;


vector< vector<int> > 
pairwise_settings( const int num_training, const int num_parameters, 
	       const vector<int> par_levels, MTRand & randgen, const int strength )
{
  vector< vector<int> > update_pair(vector< vector<int> > pair, vector<int> next_choice, const int num_parameters, vector<int>  start );
  bool check_strength( vector< vector<int> > pair, const int num_parameters, vector<int> start );
  int i, j, k ,l;
  int leftover;
  
 // Create a vector of vector<int>
  vector< vector<int> > test_settings;

   // starting number of each parameter
  vector<int>  start(num_parameters+ 1,0); 
  for( i= 0; i < num_parameters; i++)	
	start[i+1] = start[i] + par_levels[i];

  //pair matrix for strength of the pair
  vector< vector<int> > pair(start[num_parameters]);
  for(i= 0; i < start[num_parameters]; i++)	
	pair[i].resize(start[num_parameters], strength);

// First choice is random
  vector<int> temp (num_parameters);
  for ( j = 0; j < num_parameters; j++)
    temp[j] = randgen.randInt(par_levels[j] - 1);
  test_settings.push_back( temp );
  // Print the choice for debugging
  cout << "Test inputs" << endl;
  for ( k = 0; k < num_parameters; k++)
    cout << temp[k] << " ";
  cout << endl;

  pair = update_pair(pair, temp, num_parameters, start);
  
  int index;//index of  parameter chosen
  int value;// parameter value
  for ( i = 1; i < num_training; i++ )
    {
	vector<int> next_choice (num_parameters, -1);
	int best_sum = 0;
	for (j = 0; j < num_parameters; j++)
	{
		for (k = 0; k < par_levels[j]; k++)
		{
			int sum = 0;
			for ( l=0; l< start[j]; l++)
				sum += pair[l][start[j]+k];
			for (l= start[j+1]; l< start[num_parameters]; l++)
				sum += pair[start[j]+k][l];
			if (best_sum < sum)
			{
				best_sum = sum;
				value = k;
				index = j; 
			}
		}
	}
        next_choice[index] = value;
	// choosing a random order for the remaining parameters
	vector<int> random (num_parameters-1);
	for (j = 0; j < num_parameters-1; j++)
		random[j]= j;
	for (j= num_parameters-2 ; j > 1; j--)
	{
		int a = randgen.randInt(num_parameters - 2);
		int b= random[a];
		random[a]= random[j];
		random[j]= b;
	}
	for (j= 0; j< num_parameters - 1; j++)
		if (random[j] >= index)
			random[j]++;

	//choosing values for the ramaining parameters 
	for (l= 0; l< num_parameters - 1; l++)
	{
		int cur_best_sum= 0;
		for (j= 0; j< par_levels[random[l]]; j++)
		{
			int cur_sum= 0;
			for (k= 0; k< num_parameters; k++)
			{
				if (next_choice[k] != -1 )
					if (random [l]> k)
						cur_sum += pair[start[k]+next_choice[k]][start[random[l]]+j];
					else
						cur_sum += pair[start[random[l]]+j][start[k]+next_choice[k]];
			}
			if (cur_best_sum <= cur_sum)
			{
				cur_best_sum= cur_sum;
				index= random[l];
				value= j;
			}
		}
		 next_choice[index] = value;
	}
	test_settings.push_back( next_choice );

  	// Print the choice for debugging
  	for ( k = 0; k < num_parameters; k++)
   	 cout << next_choice[k] << " ";
  	cout << endl;

	pair= update_pair(pair, next_choice, num_parameters, start);
	bool done_strength= check_strength(pair, num_parameters, start);
	if (done_strength == true)
	{
/*		cout << "All pairs are covered with the specified strength in " ;
		cout << i+1 << " ";
		cout << "tests.";
  	        cout << endl;*/
                leftover = num_training - i - 1;
		i= num_training + 1;
/* old version of pairwise
		for(k= 0; k < start[num_parameters]; k++)
			for(j= 0; j < start[num_parameters]; j++)	
				pair[k][j] = strength;*/
	}
    }
    if (i == num_training)
	leftover = -1;
/*    cout << leftover << " ";
    cout << endl;*/
	if (leftover > 0 && strength == 2 )
	{
		cout << "All pairs are covered with double-pairwise in " ;
		cout << num_training - leftover << " ";
		cout << "tests.";
  	        cout << endl;
		cout << "Remaining settings will be determined by conflict heuristic." << endl;
		test_settings = limit_conflicts_settings( test_settings, leftover, 
								num_parameters,
								par_levels, randgen );
	}
	else if (leftover == 0 && strength == 2)
	{
		cout << "All pairs are covered with double-pairwise in " ;
		cout << num_training - leftover << " ";
		cout << "tests.";
  	        cout << endl;
	}
	else if (leftover < 0 && strength == 2)
	{
		test_settings.clear();
		test_settings = pairwise_settings( num_training, num_parameters,
							par_levels, randgen, 1);
	}
	else if (leftover > 0 && strength == 1)
	{
		cout << "All pairs are covered with pairwise in " ;
		cout << num_training - leftover << " ";
		cout << "tests.";
    	        cout << endl;
		cout << "Remaining settings will be determined by conflict heuristic." << endl;
		test_settings = limit_conflicts_settings( test_settings, leftover, 
								num_parameters,
								par_levels, randgen );
 	 }
	 else if (leftover == 0 && strength == 1)
	 {
		cout << "All pairs are covered with pairwise in " ;
		cout << num_training - leftover << " ";
		cout << "tests.";
   	        cout << endl;
	 }
	 else 
	 {
		cout << "Pairwise is not working. Settings will be determined by conflict heuristic." << endl;
		test_settings.clear();
		test_settings = limit_conflicts_settings( test_settings, num_training, 
								num_parameters,
								par_levels, randgen );	
       	 }
		      
  return test_settings;
}
	
//after deciding the test instance update pair matrix
vector< vector<int> > update_pair(vector< vector<int> > pair, vector<int> next_choice, const int num_parameters, vector<int>  start )
{
  for(int i =0; i < num_parameters; i++)
  {
	for (int j = i+1; j < num_parameters; j++)
	{
		if (pair[start[i] + next_choice[i]][start[j] + next_choice[j]] >= 1)
       			pair[start[i] + next_choice[i]][start[j] + next_choice[j]]--;	
	}  
  }	
  return pair;	
}

// whether all the pairs are covered with required strength
bool check_strength( vector< vector<int> > pair, const int num_parameters, vector<int> start )
{
  bool done_strength = false;
  int sum = 0;
  for(int i =0; i < num_parameters; i++)
	for (int j = start[i]; j < start[i+1]; j++)
		for (int k = start[i+1]; k < start[num_parameters]; k++)
			sum += pair[j][k];
  if (sum == 0)
	done_strength = true;
  
  return (done_strength);	
}
vector< vector<int> > 
random_settings( const int num_tests, const int num_parameters, 
	       const vector<int> par_levels, MTRand & randgen )
{
  // Design the initial set of tests
  // Create a vector of vector<int>
  vector< vector<int> > test_settings;

  // One trial is determined by a vector of integers.  
  // Use random choices for training. 
  for (int i = 0; i < num_tests; i++ )
    {
      vector<int> temp (num_parameters);

      for (int j = 0; j < num_parameters; j++)
	temp[j] = randgen.randInt(par_levels[j] - 1);

      test_settings.push_back( temp );
    }

  return test_settings;
}


vector< vector<int> > 
limit_conflicts_settings(vector< vector<int> > test_settings, const int num_tests, const int num_parameters, 
	       const vector<int> par_levels, MTRand & randgen )
{
  int i;
  // One trial is determined by a vector of integers.  

// First choice is random
  if (test_settings.size() < 1)
    {
  	vector<int> temp (num_parameters);
  	for (int j = 0; j < num_parameters; j++)
    	  temp[j] = randgen.randInt(par_levels[j] - 1);
 	test_settings.push_back( temp );
	// Print the choice for debugging
  	cout << "Test settings" << endl;
  	for (int k = 0; k < num_parameters; k++)
	  cout << temp[k] << " ";
  	cout << endl;
	i = 1;
    }
  else
    i = 0;
  for (i ; i < num_tests; i++ )
    {
      // Find a "least conflicted" set
      // This is based first on max number of overlaps, then
      // on sum of overlaps
      int best_max_conflicts = 32000; // just a big int
      int best_sum_conflicts = 32000;
      vector<int> next_choice (num_parameters); 

      bool done = false;
      vector<int> settings (num_parameters, 0);
      while (done == false)
	{
	  // compute the number in common 
	  int max = 0;
	  int sum = 0;
	  for (int j = 0; j < test_settings.size(); j++)
	    {
	      int cur_sum = 0;
	      for (int k = 0; k < num_parameters; k++)
		if (test_settings[j][k] == settings[k]) cur_sum++;

	      sum += cur_sum;
	      if (cur_sum > max)  max = cur_sum;
	    }

	  if (max < best_max_conflicts || 
	      (max == best_max_conflicts && sum < best_sum_conflicts) )
	    {
	      best_max_conflicts = max;
	      best_sum_conflicts = sum;

	      for (int k = 0; k < num_parameters; k++)
		next_choice[k] = settings[k];
	    }

	  // increment settings
	  int index = num_parameters - 1;
	  while ( (index >= 0) && (settings[index] == par_levels[index] - 1) )  
	    index--;
	  if (index == -1) 
	    done = true;
	  else
	    {
	      settings[index]++;
	      index++;
	      while (index < num_parameters)
		settings[index++] = 0;
	    }
	}

      // Print the choice for debugging
      for (int k = 0; k < num_parameters; k++)
	cout << next_choice[k] << " ";
      cout << endl;

      test_settings.push_back( next_choice );
    }

  return test_settings;
}

vector< vector<int> > 
exhaustive_settings( const int num_tests, const int num_parameters, 
		   const vector<int> par_levels )
{
  // Create a vector of vector<int>
  vector< vector<int> > test_settings;

  // One trial is determined by a vector of integers.  
  // Do all settings
  bool done = false;
  vector<int> settings (num_parameters, 0);
  while (done == false)
    {
      test_settings.push_back( settings );

      // increment settings
      int index = num_parameters - 1;
      while ( (settings[index] == par_levels[index] - 1) && (index >= 0) )  
	index--;
      if (index == -1) 
	done = true;
      else
	{
	  settings[index]++;
	  index++;
	  while (index < num_parameters)
	    settings[index++] = 0;
	}
    }
  
  // Make sure that num_tests matches what was just made
  if (num_tests != test_settings.size())
    {
      cout << "Value for num training doesn't match" << endl;
      exit(1);
    }

  return test_settings;
}

// This version starts with a candidate set, and must select from that set. 
vector< vector<int> > 
limit_conflicts_settings_subset( const int num_tests, 
				 const int num_parameters,  
				 const vector<int> par_levels, 
				 MTRand & randgen,
				 const vector< vector<int> > candidates)
{
  // Create a vector of vector<int>
  vector< vector<int> > test_settings;

  // Each trial is determined by a vector of integers.  

  // First choice is random
  test_settings.push_back( candidates[ randgen.randInt( candidates.size() - 1) ] );
  // Print the choice for debugging
  cout << "Test settings" << endl;
  for (int k = 0; k < num_parameters; k++)
    cout << test_settings[0][k] << " ";
  cout << endl;

  for (int i = 1; i < num_tests; i++ )
    {
      // Find a "least conflicted" set
      // This is based first on max number of overlaps, then
      // on sum of overlaps
      int best_max_conflicts = 32000; // just a big int
      int best_sum_conflicts = 32000;
      vector<int> next_choice (num_parameters); 

      for (int h = 0; h < candidates.size(); h++)
	{
	  // compute the number in common 
	  int max = 0;
	  int sum = 0;
	  for (int j = 0; j < test_settings.size(); j++)
	    {
	      int cur_sum = 0;
	      for (int k = 0; k < num_parameters; k++)
		if (test_settings[j][k] == candidates[h][k]) cur_sum++;

	      sum += cur_sum;
	      if (cur_sum > max)  max = cur_sum;
	    }

	  if (max < best_max_conflicts || 
	      (max == best_max_conflicts && sum < best_sum_conflicts) )
	    {
	      best_max_conflicts = max;
	      best_sum_conflicts = sum;

	      for (int k = 0; k < num_parameters; k++)
		next_choice[k] = candidates[h][k];
	    }
	}

      // Print the choice for debugging
      for (int k = 0; k < num_parameters; k++)
	cout << next_choice[k] << " ";
      cout << endl;

      test_settings.push_back( next_choice );
    }

  return test_settings;
}
